from abc import abstractmethod
from typing import Callable, Coroutine, Type
from arkitekt.messages.base import MessageModel
from enum import Enum
import asyncio
from herre import Herre
from herre.herre import get_current_herre
from koil import Koil
from koil.koil import get_current_koil
from fakts import Fakts, get_current_fakts, Config
import pydantic


class TransportConfig(Config):
    pass


class TansportConfigError(Exception):
    pass


class Transport(object):
    configClass = TransportConfig

    def __init__(
        self,
        config_dict,
        broadcast=None,
        herre: Herre = None,
        koil: Koil = None,
        fakts: Fakts = None,
        **kwargs,
    ) -> None:
        self._broadcast = broadcast
        assert (
            self._broadcast is not None
        ), "Please provide a broadcaster that receives messages"
        self.herre = herre or get_current_herre()
        self.fakts = fakts or get_current_fakts()

        try:
            self.config = self.configClass(**config_dict)
        except pydantic.error_wrappers.ValidationError as e:
            raise TansportConfigError(f"Non valid Configuration {config_dict}") from e

    @abstractmethod
    def aconnect(self):
        raise NotImplementedError()

    @abstractmethod
    def adisconnect(self):
        raise NotImplementedError()

    async def broadcast(self, message: MessageModel):
        return await self._broadcast(message)

    @abstractmethod
    async def forward(self, message: MessageModel):
        raise NotImplementedError()



from ..log import LogDataModel, LogLevel
from ....messages.types import ASSIGN_LOG
from ....messages.base import MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = ASSIGN_LOG
    extensions: Optional[MetaExtensionsModel]


class AssignLogMessage(MessageModel):
    data: LogDataModel
    meta: MetaModel

    @classmethod
    def from_log(cls, reference,  message, loglevel = LogLevel.INFO, extensions={}):
        return cls(data={
            "level": loglevel,
            "message": message
            }, meta={"extensions": extensions, "reference": reference})


from arkitekt.ui.registry import get_widget_registry
from typing import Type
from qtpy import QtWidgets, QtCore
from arkitekt.ui.qtwidgets.base import UIPortMixin
from arkitekt.schema.ports import *
from herre.wards.graphql import ParsedQuery

registry = get_widget_registry()


@registry.register(
    [(IntArgPort, SliderWidget ), (IntKwargPort, SliderWidget)]
)
class QTSliderWidget(QtWidgets.QWidget, UIPortMixin):
    port: IntKwargPort

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.selected_value = None
        self.widget: SliderWidget = self.port.widget

        self.build_ui()

    def build_ui(self):
        self.layout = QtWidgets.QVBoxLayout()

        self.sl = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.sl.setMinimum(self.widget.min or 0)
        self.sl.setMaximum(self.widget.max or 100)
        self.sl.setValue(self.port.default if isinstance(self.port, IntKwargPort) else 0)
        self.sl.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.sl.setTickInterval(5)


        self.layout.addWidget(self.sl)
        self.layout.addWidget(QtWidgets.QLabel(self.port.description))
        self.setLayout(self.layout)


    def get_current_value(self):
        return self.sl.value()






from arkitekt.schema import Node
from koil.loop import koil
from koil import get_current_koil

def ui(node: Node, *args, **kwargs):
    assert get_current_koil().state.threaded, "cannot do this in a an event loop thingy ding"
    raise NotImplementedError("Needs to be implemented Again")


from typing import Dict
from arkitekt.actors.actify import actify, define

from arkitekt.packers.transpilers.base import Transpiler
from arkitekt.schema.node import Node
from .app import AppAgent


class ScriptAgent(AppAgent):
    def register(
        self,
        *args,
        widgets={},
        transpilers: Dict[str, Transpiler] = None,
        on_provide=None,
        on_unprovide=None,
        **params
    ):
        def real_decorator(function):
            # Simple bypass for now
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            actorBuilder = actify(
                function, on_provide=on_provide, on_unprovide=on_unprovide, **params
            )

            if len(args) == 0:
                defined_node = define(function=function, widgets=widgets)
                self.templatedNewNodes.append((defined_node, actorBuilder, params))

            if len(args) == 1:
                if isinstance(args[0], str):
                    self.templatedUnqueriedNodes.append(
                        ({"q": args[0]}, actorBuilder, params)
                    )

                if isinstance(args[0], Node):
                    self.templatedNodes.append((args[0], actorBuilder, params))

                # We are registering this as a template

            return wrapped_function

        return real_decorator



from herre.wards.graphql import ParsedQuery
from arkitekt.graphql.node import DETAIL_NODE_FR


TEMPLATE_FRAGMENT = (
    """
    id
    registry {
      app {
        name
      }
      user {
        username
      }
    }
    node {
"""
    + DETAIL_NODE_FR
    + """
    }
"""
)


TEMPLATE_GET_QUERY = ParsedQuery(
    """
query Template($id: ID,){
  template(id: $id){
    """
    + TEMPLATE_FRAGMENT
    + """
  }
}
"""
)


UPDATE_OR_CREATE_TEMPLATE = ParsedQuery(
    """
mutation CreateTemplate($node: ID!, $params: GenericScalar, $extensions: [String], $version: String){
  createTemplate(node: $node, params: $params, extensions: $extensions, version: $version){
    """
    + TEMPLATE_FRAGMENT
    + """
  }
}
"""
)



from arkitekt.transport.base import Transport, TransportConfig
import logging

logger = logging.getLogger(__name__)


class AgentTransportConfig(TransportConfig):
    identifier: str = "main"

    @property
    def protocol(self):
        return "wss" if self.secure else "ws"


class AgentTransport(Transport):
    configClass = AgentTransportConfig



from arkitekt.packers.structure import Structure, StructureMeta
from pydantic import BaseModel
import inspect
from arkitekt.schema.widgets import Widget


def props(obj):
    pr = {}
    for name in dir(obj):
        value = getattr(obj, name)
        if not name.startswith('__') and not inspect.ismethod(value):
            pr[name] = value
    return pr


class StructureModel(Structure):
    """Model

    Model is the abstract baseclass of all Serverside Models and provides a Django ORM
    like interface for retrieving data from the Server.

    Implements:
        id (str): Every Model has an id (UUID) that identifies the Server Instance

    Args:
        BaseModel ([type]): [description]
        Structure ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ModelMeta.

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """


    @classmethod
    def get_structure_meta(cls) -> StructureMeta:
        meta = props(cls.get_meta())
        print(meta)
        return StructureMeta(**meta)



from abc import ABC, abstractclassmethod, abstractmethod
from typing import AbstractSet, Union

from pydantic.main import BaseModel

from arkitekt.schema.widgets import Widget
from enum import Enum

class BoundType(str, Enum):
    GLOBAL = "GLOBAL"
    AGENT = "AGENT"
    REGISTRY = "REGISTRY"
    APP = "APP"



class StructureMeta(BaseModel):
    widget: Widget = None
    identifier: str
    bound: BoundType = BoundType.GLOBAL
    overwrite = False


class Structure(ABC):
    """An Abstract Mixin to enforce packaging as a Structure before registering

    Args:
        ABC ([type]): [description]

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """


    @abstractclassmethod
    def get_structure_meta(cls) -> StructureMeta:
        return StructureMeta()

    @abstractclassmethod
    async def expand(cls, shrinked_value):
        raise NotImplementedError("Every Structure needs to implement a expand Method")

    @abstractmethod
    async def shrink(self):
        raise NotImplementedError("Every Structure needs to implement a shrink Method")



from ..log import LogDataModel
from enum import Enum
from pydantic.main import BaseModel
from ....messages.types import PROVIDE_TRANSITION
from ....messages.base import (
    MessageDataModel,
    MessageMetaExtensionsModel,
    MessageMetaModel,
    MessageModel,
)
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]


class MetaModel(MessageMetaModel):
    type: str = PROVIDE_TRANSITION
    extensions: Optional[MetaExtensionsModel]


class ProvideState(str, Enum):
    # Start State
    PENDING = "PENDING"
    PROVIDING = "PROVIDING"

    # Life States
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    CANCELING = "CANCELING"
    DISCONNECTED = "LOST"
    RECONNECTING = "RECONNECTING"
    # End States
    DENIED = "DENIED"

    # End States
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"
    ENDED = "ENDED"
    CANCELLED = "CANCELLED"


class ProvideMode(str, Enum):
    # Start State
    DEBUG = "DEBUG"
    PRODUCTION = "PRODUCTION"


class ProvideTransistionData(MessageDataModel):
    state: ProvideState
    mode: ProvideMode = ProvideMode.PRODUCTION
    message: Optional[str]


class ProvideTransitionMessage(MessageModel):
    data: ProvideTransistionData
    meta: MetaModel

    @classmethod
    def from_critical(cls, reference, exception, extensions={}):
        return cls(
            data={"message": str(exception), "state": ProvideState.CRITICAL},
            meta={"extensions": extensions, "reference": reference},
        )



from arkitekt.actors.exceptions import ThreadedActorCancelled
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.legacy.utils import get_running_loop
from arkitekt.threadvars import get_current_assign, get_current_cancel_event, get_current_transport, get_current_janus
from arkitekt.messages.postman.log import LogLevel




async def log_async(message, level: LogLevel = LogLevel.INFO):
    transport = get_current_transport()
    assign = get_current_assign()

    message = AssignLogMessage(
        data = {
            "message": str(message),
            "level": level
        },
        meta = {**assign.meta.dict(exclude={"type"})}
    )

    await transport.forward(message)


def useUser():
    assign = get_current_assign()
    return assign.meta.context.user

def cancelCheck():
    event = get_current_cancel_event()
    if event.is_set(): raise ThreadedActorCancelled("Actor Was Cancelled")


def log(message: str, level: LogLevel = LogLevel.INFO):
    """ Logs a message

    Depending on both the configuration of Arkitekt and the overwrite set on the
    Assignment, this logging will be sent (and persisted) on the Arkitekt server

    Args:
        message (sr): The Message you want to send
        level (DebugLevel, optional): The level of the log. Defaults to DebugLevel.INFO.

    Returns:
        [Future]: Returns a future if currently running in an event loop
    """
    try:
        event_loop = get_running_loop() # Check if we are in an event loop
    except RuntimeError:
        sync_q = get_current_janus()
        sync_q.put(("log", (message, level)))
        sync_q.join()
    else:
        return log_async(message, level=level)


from .unprovide_critical import UnprovideCriticalMessage
from .unprovide_log import UnprovideLogMessage
from .bounced_unprovide import BouncedUnprovideMessage
from .unprovide import UnprovideMessage
from .bounced_unprovide import BouncedUnprovideMessage
from .unprovide_done import UnprovideDoneMessage
from .unprovide_error import UnprovideErrorMessage